@@ -10,7 +10,9 @@
 
 module Main where
 
+#if !MIN_VERSION_base(4,8,0)
 import           Control.Applicative
+#endif
 import           Control.Monad
 import           Data.Int
 import           Data.Word
@@ -23,17 +25,21 @@ import qualified Test.SmallCheck.Series             as SC
 import qualified Test.Tasty.SmallCheck              as SC  (testProperty)
 import qualified Test.QuickCheck                    as QC
 import qualified Test.Tasty.QuickCheck              as QC  (testProperty)
+import qualified Data.Binary                        as Binary (encode, decode)
 import qualified Data.Text.Lazy                     as TL  (unpack)
 import qualified Data.Text.Lazy.Builder             as TLB (toLazyText)
+import qualified Data.Text.Lazy.Builder.Scientific  as T
+
+#ifdef BYTESTRING_BUILDER
 import qualified Data.ByteString.Lazy.Char8         as BLC8
 import qualified Data.ByteString.Builder.Scientific as B
-import qualified Data.Text.Lazy.Builder.Scientific  as T
 
 #if !MIN_VERSION_bytestring(0,10,2)
 import qualified Data.ByteString.Lazy.Builder       as B
 #else
 import qualified Data.ByteString.Builder            as B
 #endif
+#endif
 
 main :: IO ()
 main = testMain $ testGroup "scientific"
@@ -43,6 +49,11 @@ main = testMain $ testGroup "scientific"
        (QC.forAll normalizedScientificGen    $ \s ->
             s /= 0 QC.==> abs (Scientific.coefficient s) `mod` 10 /= 0)
 
+  , testGroup "Binary"
+    [ testProperty "decode . encode == id" $ \s ->
+        Binary.decode (Binary.encode s) === s
+    ]
+
   , testGroup "Parsing"
     [ testCase "reads \"\""        $ testReads ""        []
     , testCase "reads \"1.\""      $ testReads "1."      [(1.0, ".")]
@@ -61,16 +72,20 @@ main = testMain $ testGroup "scientific"
 
     , testGroup "Builder"
       [ testProperty "Text" $ \s ->
-          formatScientific B.Generic Nothing s ==
-          TL.unpack (TLB.toLazyText $ T.formatScientificBuilder B.Generic Nothing s)
+          formatScientific Scientific.Generic Nothing s ==
+          TL.unpack (TLB.toLazyText $
+                       T.formatScientificBuilder Scientific.Generic Nothing s)
 
+#ifdef BYTESTRING_BUILDER
       , testProperty "ByteString" $ \s ->
-          formatScientific B.Generic Nothing s ==
-          BLC8.unpack (B.toLazyByteString $ B.formatScientificBuilder B.Generic Nothing s)
+          formatScientific Scientific.Generic Nothing s ==
+          BLC8.unpack (B.toLazyByteString $
+                        B.formatScientificBuilder Scientific.Generic Nothing s)
+#endif
       ]
 
     , testProperty "formatScientific_fromFloatDigits" $ \(d::Double) ->
-        formatScientific B.Generic Nothing (Scientific.fromFloatDigits d) ==
+        formatScientific Scientific.Generic Nothing (Scientific.fromFloatDigits d) ==
         show d
 
     -- , testProperty "formatScientific_realToFrac" $ \(d::Double) ->
@@ -141,7 +156,15 @@ main = testMain $ testGroup "scientific"
     ]
 
   , testGroup "Conversions"
-    [ testGroup "Float"  $ conversionsProperties (undefined :: Float)
+    [ testProperty "fromRationalRepetend" $ \(l, r) -> r ==
+        (case fromRationalRepetend (Just l) r of
+          Left (s, rr) -> toRational s + rr
+          Right (s, mbRepetend) ->
+            case mbRepetend of
+              Nothing       -> toRational s
+              Just repetend -> toRationalRepetend s repetend)
+
+    , testGroup "Float"  $ conversionsProperties (undefined :: Float)
     , testGroup "Double" $ conversionsProperties (undefined :: Double)
 
     , testGroup "floatingOrInteger"
@@ -342,10 +365,13 @@ normalizedScientificSeries = liftM Scientific.normalize SC.series
 
 instance QC.Arbitrary Scientific where
     arbitrary = QC.frequency
-                  [ (70, scientific <$> QC.arbitrary <*> intGen)
-                  , (20, scientific <$> QC.arbitrary <*> bigIntGen)
-                  , (10, scientific <$> pure 0       <*> bigIntGen)
-                  ]
+      [ (70, scientific <$> QC.arbitrary
+                        <*> intGen)
+      , (20, scientific <$> QC.arbitrary
+                        <*> bigIntGen)
+      , (10, scientific <$> pure 0
+                        <*> bigIntGen)
+      ]
 
     shrink s = zipWith scientific (QC.shrink $ Scientific.coefficient s)
                                   (QC.shrink $ Scientific.base10Exponent s)
